#
# estcore.rb
#
#     Copyright(c) 1999 Minero Aoki
#     aamine@dp.u-netsurf.ne.jp
#

require 'nkf'
require 'thread'
require 'delegate'
require 'singleton'

require 'must'
require 'extmod'
require 'gconst'
require 'bug'
require 'tmail'
require 'loader'
require 'smtpclient'
require 'popclient'

require 'est/readrc'


module EST

  Version = '0.0.2'


  class CCore

    include Singleton

    private

    def do_initialize
      init_rc

      @composer = Composer.new
      @sender   = Sender.new
      @popper   = Popper.new
      @mailset  = Mailset.new
    end

    def init_rc
      # @rc = ESTrcRoot.new
      @rc = setrc( File.expand_path '~/.estrc' )

      require 'est/estconf'
      @rc['core'] = estconfig()
    end


    def chk( tag, val )
      # puts "#{tag}: #{val.inspect}"
    end

    def setrc( fname )
      p = ESTrcp.new
      ret = p.parse( fname )

      #-------------------------------------------------------
      # data
      #

      kconfig    = ret['config']
      ksender    = ret['sender']['server']
      kpopper    = ret['popper']['server']
      kloader    = ret['mailbox']['mbox']
      kcomposer  = ret['composer']['entry']
      kinterface = ret['interface'][ File.basename $0 ]
      krefile    = ret['mailbox']['refile']

      chk 'kconfig', kconfig
      chk 'ksender', ksender
      chk 'kpopper', kpopper
      chk 'kloader', kloader
      chk 'kcomposer', kcomposer
      chk 'kinteface', kinterface
      chk 'krefile', krefile

      #-------------------------------------------------------
      # profile
      #

      conf = kconfig[ ret[ 'use' ] ]

      vsender = ksender[ conf[ 'sender' ] ]
      vsender['port'] = vsender['port'].to_i
      
      vpopper = {}
      conf[ 'popper' ].split(',').each do |srv|
        srv.strip!
        phash = {}
        kpopper[ srv ].each do |k,v|
          case k
          when 'port'
            v = v.to_i
          end
          phash[ k ] = v
        end
        vpopper[ srv ] = phash
      end
      vloader   = ret[ 'mailbox' ][ 'system' ][ conf[ 'mailbox' ] ]
      vcomposer = kcomposer[ conf[ 'composer' ] ]
      vrefile   = mkrefilearg( krefile )

      chk 'conf', conf
      chk 'vsender',   vsender
      chk 'vpopper',   vpopper
      chk 'vloader',   vloader
      chk 'vcomposer', vcomposer
      chk 'vrefile',   vrefile

      vl = {}
      (kloader.keys - vloader.values).each do |name|
        vl.store name, kloader[name]
      end
      vloader.each do |name,real|
        vl.store name, kloader[real]
      end
      vloader = vl

      #-------------------------------------------------------
      # set
      #

      ret = {
        'current' => {
          'sender' => vsender,
          'popper' => vpopper,
          'mailbox' => vloader,
          'composer' => vcomposer
        },
        'sender' => ksender,
        'popper' => kpopper,
        'mailbox' => kloader,
        'composer' => kcomposer,
        'if' => kinterface
      }

      return ret
    end


    def mkrefilearg( arg )
      return nil unless arg

      arr = []
      arg.each do |num, str|
        # "str" is like [ header, refile_regex, dest_mbox ]
        arr[ num.to_i ] = eval( str )
      end
      arr.compact!

      return arr
    end



    public


    attr :rc
    attr :mailset
    attr :composer
    attr :sender
    attr :popper

    def []( key ) @rc['current'][ key ] end

    def notice( str )
      $stderr.puts str
    end
    
    def shutdown
      puts 'EST core.shutdown ...'
      @mailset.shutdown if @mailset
      puts 'done'
    end

		def load_cache( name )
			nil
		  # @rc['core']['libdir']
end  # @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

  end

  class Core < CCore
  end



  class CMailset

    def initialize
      @set = {}
    end

    def shutdown
    end


    def []( name )
      name.must String   #D#

      unless ret = @set[ name ] then
        if rc = $CORE['mailbox'][name] then
          lo = mkloader( rc )
          ret = realizembox( name, rc, lo )
          @set.store( name, ret )
        else
          ret = nombox( name )
        end
      end

      return ret
    end


    def mkloader( rc )
      case rc[ 'type' ]
      when 'mh'
        lo = MhLoader.new( rc['path'] )
      else
        bug! 'mbox type is not MH'
      end

      return lo
    end

    def realizembox( name, rc, lo )
      Mailbox.new( name, rc, lo )
    end

    def nombox( name )
      raise NameError, "no such mailbox: #{name}"
    end

    def get_targetmbox( name )
      self[ name ]
    end


    def mvmail( from_name, idx, to_name )
      return if from_name == to_name
      return unless from = self[ from_name ]
      return unless to = get_targetmbox( to_name )
      return unless port = from[idx]

      to.mv port
      from.rm idx
    end


    def cpmail( idx, from_name, to_name )
      return unless from = self[ from_name ]
      return unless to = get_targetmbox( to_name )
      return unless port = from[idx]

      to.cp port
    end


    def rmmail( box_name, idx )
      return unless mbox = self[ box_name ]
      mbox.rm idx
    end


    def mkmail( box_name, str = nil )
      return nil unless mbox = self[ box_name ]

      port = mbox.new_port
      port.wopen do |wp|
        if str then
          wp.write str
        else
          yield wp
        end
      end

      mbox.add port

      return port
    end


    def refile
      to_mb = nil
      inbox = self['inbox']

      inbox.each_with_index do |port, idx|
        to_mb = nil
        nil.each do |arr|
          nm, re, to = *arr
          mail = Mail.new( port )

          if mail[nm] =~ re then
            to_mb = to
            break
          end
        end
        mvmail( 'inbox', idx, to_mb ) if to_mb
      end
    end

  end

  class Mailset < CMailset
  end



  class CMailbox

    include Enumerable


    private

    def initialize( name, rc, lo )
      name.must String
      @name = name.dup.freeze
      @rc = rc
      @loader = lo
      @array = []
    end

    def do_initialize
    end

    abstract :do_add
    abstract :do_cp
    abstract :do_mv
    abstract :do_rm
    abstract :do_new_port


    public

    def add( m )
      do_add m
    end
        
    def cp( port )
      do_cp port
    end

    def mv( port )
      do_mv port
    end

    def rm( idx )
      do_rm idx
    end

    def new_port
      do_new_port
    end

    
    def update
    end

    attr :name

    def []( idx )
      return nil unless idx
      @array[idx]
    end

    def size
      @array.size
    end
    alias length size

    
    def each
      @array.each{|i| yield i }
    end

    def each_index
      @array.each_index{|i| yield i }
    end

    def each_with_index
      @array.each_with_index{|item,idx| yield item,idx }
    end

    def index( port )
      port.must Port
      @array.index port
    end

    def rindex( port )
      port.must Port
      @array.rindex port
    end

    def indexes( *args )
      @array.indexes *args
    end

    def sort
      @array.sort!{|a,b| yield a,b }
    end

    def to_a
      @array.dup
    end

  end

  class Mailbox < CMailbox
  end



  class RealMailbox < Mailbox

    def update
      hash = {}
      each{|port| hash[ port ] = true }
      @loader.each do |p|
        p.unread = true if @name == 'inbox'
        add p unless hash.has_key? p
      end
    end


    private

    def do_initialize
			unread = (@name == 'inbox')

			if cache = $CORE.load_cache( @name ) then
			  cache.each do |port, ifdata|
					port.listcache = ifdata
					port.unread = unread
					add p
				end
			else
				@loader.each do |p|
					p.unread = unread
					add p
				end
			end
    end

    def do_add( port )
      port.must Port
      @array.push port
      port.ref

      return port
    end

    def do_cp( port )
      port.must Port

      p = @loader.new_port
      port.cp p
      p.unread = port.unread
      add p

      return p
    end

    def do_mv( port )
      port.must Port

      p = @loader.new_port
# puts "do_mv: from=#{port},to=#{p}"
      port.mv p
      p.unread = port.unread
      add p

      return p
    end

    def do_rm( idx )
      return unless port = @array.delete_at( idx )
      port.unref
      return nil
    end

    def do_new_port
      return @loader.new_port
    end
  
  end


  class ReferenceError < StandardError ; end


  class VirtualMailbox < Mailbox

    private

    def do_add( port )
      port.must Port
      @array.push port
      port.ref
    end

    def do_cp( port )
      raise ReferenceError, "cannot copy port to virtual mbox"
    end

    def do_mv( port )
      raise ReferenceError, "cannot move port to virtual mbox"
    end

    def do_rm( idx )
      port = @array.delete_at( idx )
      port.unref if port
    end

    def do_new_port
      raise ReferenceError, "virtual mbox cannot make connection"
    end

  end



  class Mail < TTMail

    include TMailEasyAccess

    def initialize( port )
      port.must Port
      super
    end

    def port
      @port
    end

    def port=( p )
      unless @parsed then
        body
      end
      @port.mv p
      @port = p
    end
      
    def sync
      @port.wopen do |wp|
        wp.write to_s
      end
    end

  end



  class CPopper

    def load
      $CORE['popper'].each do |name, rc|
        do_pop( name, rc )
      end
    end


    def do_pop( name, rc )
      client = nil
      addr = rc[ 'address' ]
      port = rc[ 'port' ]
      acnt = rc[ 'account' ]
      pass = rc[ 'password' ]
      if_nopassword( acnt, pass ) unless pass

      cls = POP3client
      client = cls.new( addr, port )
      if_loggingin( client )
      begin
        client.login( acnt, pass )
        if_loggedin( client )
      ensure
        client.logout
      end
      if_loggedout( client )
    end

    def if_nopassword( acnt, pass )
      raise ArgumentError, "POP password is not given"
    end

    def if_loggingin( client )
    end

    def if_loggedin( client )
      client.each do |popmail|
        port = $CORE.mailset.mkmail( 'inbox' ){|wp|
          popmail.mail wp
        }
        if_popmail port
        popmail.delete
      end
    end

    def if_popmail( port )
      port.unread = true
    end

    def if_loggedout( client )
    end

  end

  class Popper < CPopper
  end



  class CSender

    def sendmail
      if $CORE.mailset['draft'].size == 0 then
        if_nosendmail
        return
      end

      client = nil
      arg = $CORE['sender']

      addr = arg['address']
      port = arg['port']
      helo = arg['helo-domain']

      cls = SMTPclient
      client = cls.new( addr, port )
      if_loggingin( client )
      begin
        client.login( helo )
        if_loggedin( client )
      ensure
        client.logout
      end
      if_loggedout( client )
    end


    def if_nosendmail
    end

    def if_loggingin( client )
    end

    def if_loggedin( client )
      $CORE.mailset['draft'].each_with_index do |port,idx|
        mail = Mail.new( port )
        fromaddr = $CORE.composer.send_fromaddr( mail, client )
        toaddrs  = $CORE.composer.send_toaddrs( mail, client )
        mailstr  = $CORE.composer.send_text( mail, client )
        client.sendmail( mailstr, fromaddr, toaddrs )
        if_sendmail( mail, idx )
      end
    end

    def if_loggedout( client )
      return unless tomb = $CORE['sender']['moveto']

      (0 .. $CORE.mailset['draft'].size - 1).each do |i|
        $CORE.mailset.mvmail( 'draft', i, tomb )
      end
    end

    def if_sendmail( mail, idx )
    end

  end


  class Sender < CSender
  end



  class CComposer

    def initialize
    end


    def send_text( mail, client )
      mail.body = Nkf.nkf( '-j', mail.body )
      mail.to_s( "\r\n" )
    end

    def send_fromaddr( mail, client )
      faddr = mail.from
      unless faddr then
        raise ArgumentError, "no 'from' address designation in mail"
      end

      return faddr
    end

    def send_toaddrs( mail, client )
      toaddr = []
      mail.each_dest do |addr|
        toaddr.push addr
      end
      if toaddr.size == 0 then
        raise ArgumentError,
          "no 'to' address designation in mail"
      end

      return toaddr
    end


    REPLYTOHEADER = {
      'To'          => %|replyto_to(original)|,
      'Subject'     => %|replyto_subject(original)|,
      'In-Reply-To' => %|original.msgid|,
      'References'  => %|original.msgid|
    }

    def replyto( original )
      case original
      when Mail
        ;
      when Port
        original = Mail.new( original )
      else
        original.must Mail, Port
      end

      port = $CORE.mailset.mkmail( 'draft' ){|wp|
        NEWHEADER.dup.update( REPLYTOHEADER ).each do |n,v|
          wp.puts( n + ': ' + eval(v) )
        end
        wp << "\n"
        wp << replyto_body( original )
      }
      return port
    end


    def replyto_to( mail )
      if hed = mail['reply-to']        then return hed.body end
      if hed = mail['from']            then return hed.body end
      if hed = mail['resent-reply-to'] then return hed.body end
      if hed = mail['resent-from']     then return hed.body end
      return ''
    end

    
    def replyto_subject( orig )
      case sbj = orig.subject
      when /\A\s*\[[^\]]+\]\s*re\:\s*(\S.*)/io,
           /\A\s*\[[^\]]+\]\s*(\S.*)/io
        sbj = 'Re: ' + $1
      when /\A\s*re:\s*(\S+)/io === sbj then
        ;
      else
        sbj += 'Re: '
      end

      return sbj
    end

    
    def replyto_body( mail )
      str = "\n\n  In mail \""
      str << mail.subject('') << "\"\n  "
      str << (mail.from_phrase('') || mail.from(''))
      str << " wrote:\n\n"
      mail.body.each do |line|
        str << '> ' << line
      end

      return str
    end


    NEWHEADER = {
      'Date'         => %|"\#{DateH.t2s(Time.now)}"|,
      'Subject'      => %|""|,
      'From'         => %|$CORE['composer']['from-address']|,
      'MIME-Version' => %|"1.0"|,
      'Content-Type' => %|"text/plain; charset=iso-2022-jp"|,
      'X-Mailer'     => %|"Mailest version \#{EST::Version}"|
    }

    def new_mail
      port = $CORE.mailset.mkmail( 'draft' ){|wp|
        NEWHEADER.each do |n,v|
          wp.puts n + ': ' + eval(v)
        end
        wp.puts "\n\n"
      }
      return port
    end


    def list_items( m )
      [ list_date( m['date'] && m['date'].date ),
        list_from( m ),
        list_subject( m.subject ) ]
    end

    def list_date( ti )
      if ti
        ti.strftime( '%m/%d %H:%M' )
      else
        ''
      end
    end

    def list_from( tm )
      tm.from_phrase(nil) || tm.from('')
    end

    def list_subject( str )
      if /\A\s*\[[^\]]+\]\s*/o === str then
        return $'
      else
        return str
      end
    end


    HEADER_ALIST = [
     'date',
     'from',
     'subject',
     'x-mailer'
    ]

    def view_upper( mail )
      arr = []
      mail.each do |k,v|
        unless HEADER_ALIST.include? k then
          arr.push v.inspect
        end
      end
      return arr.join("\n")
    end


    def view_lower( mail )
      tmp = HEADER_ALIST.collect{|k| mail[k].inspect if mail[k] }
      return tmp.compact.join("\n") << "\n\n" << mail_to_viewtext( mail ) 
    end


    def mail_to_viewtext( mail )
      ret = nil
      mtype = mail.maintype.downcase

      case mtype
      when 'text', ''
        ret = mail.body
        case mail.encoding.downcase
        when '7bit', '8bit', ''
          ;
        when 'base64'
          ret = ret.unpack('m')[0]
        else
          return undecodable_mail( mail )
        end
      when 'multipart'
        mail.parts.each do |m|
          tmp = m.maintype.downcase
          if tmp == 'text' or tmp == '' then
            return mail_to_viewtext(m)
          end
        end
        return unknown_type( mail )
      else
        return unknown_type( mail )
      end

      if NKF.guess( ret ) != NKF::EUC then
        ret = NKF.nkf( '-e', ret )
      end

      # search for "\r\n" from first 200 char
      from = ret.size - 200
      from = from > 0 ? from : 0
      if ret.index( "\r", from ) then
        ret = ret.gsub( "\r\n", "\n" )
      end

      return ret
    end

    def undecodable_mail( mail )
      mail['content-transfer-encoding'].inspect
    end

    def unknown_type( mail )
      mail['content-type'].inspect
    end

    def standby_send( mail )
      mail.to_s
    end

  end

  class Composer < CComposer
  end


end



# post-define Port (in 'port.rb')

class Port

  def initialize
    @refc = 0
    @unread = false
    @listcache = nil
  end

  attr :unread, true
  attr :listcache, true

  def ref
    @refc += 1
  end

  def unref
    @refc -= 1
puts "port#unref, refc=#{@refc}"
    if @refc < 1 then
puts "delete file"
      self.rm
    else
puts "not delete"
    end
  end

end
