#
# mailest graphical user interface using Gtk+
#
#     Copyright(c) 1999 Minero Aoki
#     aamine@dp.u-netsurf.ne.jp
#

require 'est/estcore'
require 'gtk'


module EST

  class Core

    private

    def initialize
      @editers       = []
      @working       = false
      @shutting_down = false
    end

    def init_rc
      super
      @rc['if']['libdir'] = @rc['core']['libdir'] + '/' + 'mailest'
    end


    public

    attr :working

    def trigger
      return false if @working
      @working = true

      do_initialize
      @mailset.init_after

      notice "mailest ready"
      Gtk.main
    end


    def edit( port )
      unless @shutting_down then
        rc = $CORE.rc['if']['editer']
        cls = (rc['type'] == 'internal') ? InternalEditer : ExternalEditer
        @editers.push cls.new( port, rc )
        return true
      else
        return false
      end
    end

    def edit_done( ed )
      unless @shutting_down then
        @editers.delete( ed )
      end
    end

    def editers
      @editers.dup
    end

    def shutdown( force = false )
      return if @shutting_down
      @shutting_down = true
      unless force then
        unless quit_ok? then
          @shutting_down = false
          return
        end
      end

      # shutdown realy
      notice 'going to shutdown...'
      @editers.each do |ed| ed.shutdown end
      @editers.clear
      super()
      exit 0
      @shutting_down = false    # not reach
    end

    def notice( str )
      mailset.write str
    end


    private

    def quit_ok?
      true
    end

  end   # class Core



  class Mailset

    def initialize
      super

      make_win
      pack_win
      create_pm
      init_event
      @inited = false
    end

    def init_after
      return if @inited
      @inited = true

      @current = self['inbox']
      @win.show_all
      @current.select 0
    end


    def shutdown
      @win.destroy
      super
    end


    def view( mail )
      upper = $CORE.composer.view_upper( mail )
      lower = $CORE.composer.view_lower( mail )

      @text.freeze
      @text.set_point 0
      @text.forward_delete @text.get_length

      @text.insert_text( upper, 0 )
      @text.thaw
      val = @textadj.upper

      @text.freeze
      @text.insert_text( "\n", @text.get_length )
      @text.insert_text( lower, @text.get_length )
      @text.thaw
      @textadj.value = val
    end


    def write( str )
      @notice.set str.to_s
    end


    def chmbox( name )
      name.strip!
      return if @current.name == name
      return unless tmp = get_targetmbox( name )

      # keep mbox single select mode
      if @current.multiselect then
        @excheck.clicked
      end

      old = @current
      @current = tmp
      @current.show
      old.hide
      
      @current.select @current.selected

      if @current.size == 0 then
        $CORE.notice "no mail"
      else
        $CORE.notice "#{@current.size} mails"
      end
    end


    def discard_mail( mbox, idx )
      if mbox.name == 'trash' then
        rmmail( 'trash', idx )
      else
        mvmail( mbox.name, idx, 'trash' )
      end
      if idx >= mbox.size then
        idx = mbox.size - 1
      end
      mbox.select idx
    end


    def read_ahead
      # 90% ahead
      val = @textadj.value + @textadj.page_size * 9 / 10

      if val >= @textadj.upper then   # at tail
        @current.view_next
      else
        @textadj.value = [ val, @textadj.upper - 1 ].min
      end
    end


    def read_back
      if @textadj.value <= @textadj.lower then   # at head
        @current.view_prev
      else
        # 90% back
        @textadj.value = [ @textadj.value - @textadj.page_size * 9 / 10,
                           @textadj.lower ].max
      end
    end



    private


    def make_win
      @win         = Gtk::Window.new( Gtk::WINDOW_TOPLEVEL )
      @win.set_title( 'mailest' )
      @win.border_width(0)
      @win.set_usize( 640, 480 )

      @bigbox      = Gtk::VBox.new( false, 0 )
      @topbox      = Gtk::HBox.new( false, 0 )
      @listbox     = Gtk::HBox.new( false, 0 )
      @clistbox    = Gtk::VBox.new( false, 0 )

      @mboxcombo   = Gtk::Combo.new
      @mboxcombo.set_usize( 100, 24 )
      @mboxcombo.entry.set_editable false
      @mboxcombo.set_popdown_strings $CORE['mailbox'].keys
      @mboxcombo.entry.set_text 'inbox'

      @excheck     = Gtk::CheckButton.new('')

      mrc = $CORE.rc['if']['menu']
      @menubar     = Gtk::MenuBar.new
      @menumove    = Gtk::Menu.new
      @itemmove    = Gtk::MenuItem.new( mrc['move'] )
      @itemtrash   = Gtk::MenuItem.new( mrc['trash'] )
      @itemmbox    = Gtk::MenuItem.new( mrc['mbox'] )
      @menumbox    = Gtk::Menu.new
      @mitemupdate = Gtk::MenuItem.new( mrc['update'] )
      @mitemclear  = Gtk::MenuItem.new( mrc['clear'] )
      @mitemahead  = Gtk::MenuItem.new( mrc['ahead'] )
      @mitemback   = Gtk::MenuItem.new( mrc['back'] )
      @mitemnext   = Gtk::MenuItem.new( mrc['next'] )
      @mitemprev   = Gtk::MenuItem.new( mrc['prev'] )
      @itemmail    = Gtk::MenuItem.new( mrc['mail'] )
      @menumail    = Gtk::Menu.new
      @mitemwrite  = Gtk::MenuItem.new( mrc['write'] )
      @mitemreply  = Gtk::MenuItem.new( mrc['reply'] )
      @mitemreedit = Gtk::MenuItem.new( mrc['reedit'] )
      @mitemfwd    = Gtk::MenuItem.new( mrc['forward'] )
      @itempop     = Gtk::MenuItem.new( mrc['pop'] )
      @itemsend    = Gtk::MenuItem.new( mrc['send'] )
      @itemquit    = Gtk::MenuItem.new( mrc['quit'] )

      @paned       = Gtk::VPaned.new
      @paned.handle_size 7
      @paned.gutter_size 7

      @textbox     = Gtk::VBox.new( false, 0 )

      @notice      = Gtk::Label.new( 'Copyright (c) 1999 Minero Aoki' )
      @notice.set_usize( 640, 20 )
      @notice.set_alignment( 0, 0.5 )

      @textadj     = Gtk::Adjustment.new( 0, 0, 0, 0, 0, 0 )
      @textscr     = Gtk::VScrollbar.new( @textadj )
      #  @text        = Gtk::Text.new( nil, @textadj )
      @text        = Gtk::Buffer.new( nil, @textadj )
      @text.set_usize( 640, 200 )
      @text.set_editable false

      @accel       = Gtk::AccelGroup.new
      @accel.attach @win
      @accel.add( ?\ , 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemahead, 'activate' )
      @accel.add( ?j, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemahead, 'activate' )
      @accel.add( ?k, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemback, 'activate' )
      @accel.add( ?n, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemnext, 'activate' )
      @accel.add( ?p, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemprev, 'activate' )
      @accel.add( ?w, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemwrite, 'activate' )
      @accel.add( ?a, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemreply, 'activate' )
      @accel.add( ?e, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemreedit, 'activate' )
      @accel.add( ?f, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @mitemfwd, 'activate' )
      @accel.add( ?q, Gdk::SHIFT_MASK, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @itemquit, 'activate' )
    end

    def pack_win
      @win.add( @bigbox )
      @bigbox.pack_start( @topbox, false, false, 0 )
      @bigbox.pack_start( @paned, true, true, 0 )

      @topbox.pack_start( @excheck, false, false, 0 )
      @topbox.pack_start( @mboxcombo, false, false, 0 )
      @topbox.pack_start( @menubar, true, true, 0 )

      @paned.add1 @listbox
      @paned.add2 @textbox

      @listbox.pack_start( @textscr, false, false, 0 )
      @listbox.pack_start( @clistbox, true, true, 0 )

      @menubar.append @itemmove
      @menubar.append @itemtrash
      @menubar.append @itemmbox
      @menubar.append @itemmail
      @menubar.append @itempop
      @menubar.append @itemsend
      @menubar.append @itemquit
      @itemmove.set_submenu @menumove
      @itemmbox.set_submenu @menumbox
      @menumbox.add @mitemupdate
      @menumbox.add @mitemclear
      # @menumbox.add @mitemahead
      # @menumbox.add @mitemback
      @menumbox.add @mitemnext
      @menumbox.add @mitemprev
      @itemmail.set_submenu @menumail
      @menumail.add @mitemreply
      @menumail.add @mitemwrite
      @menumail.add @mitemreedit
      @menumail.add @mitemfwd

      @textbox.pack_start( @notice, false, false, 0 )
      @textbox.pack_start( @text, true, true, 0 )
#D#=begin
      @itemgc = Gtk::MenuItem.new( 'GC' )
      @itemgc.signal_connect( 'activate' ) do |w|
        GC.start
      end
      @menubar.append @itemgc
#D#=end
    end


    def create_pm
      @win.realize
      libdir = $CORE.rc['if']['libdir'] + '/'
      @mailpm, @mailmask = Gdk::Pixmap.create_from_xpm(
        @win.window, nil, libdir + 'whitemail.xpm' )
      @mimepm, @mimemask = Gdk::Pixmap.create_from_xpm(
        @win.window, nil, libdir + 'mimemail.xpm' )
    end


    def init_event
      @win.signal_connect( 'destroy' ) do |w|
        $CORE.shutdown( true )
        true
      end

      $CORE['mailbox'].keys.each do |mbname|
        mitem = Gtk::MenuItem.new( mbname )
        mitem.signal_connect( 'activate' ) do |w|
protect( 'move' ){    #D#
          @current.each_selection do |idx|
            mvmail( @current.name, idx, mbname )
          end
}    #D#
        end
        @menumove.add mitem
      end

      @excheck.signal_connect( 'toggled' ) do |w|
        @current.multiselect = ! @current.multiselect
      end

      @itemtrash.signal_connect( 'activate' ) do |w|
protect( 'poi' ){    #D#
        c = 0
        @current.each_selection do |idx|
          c += 1
          discard_mail( @current, idx )
        end
        case c
        when 0
          $CORE.notice "no mail selected"
        when 1
          ;
        else
          $CORE.notice "discard #{c} mails"
        end
}    #D#
      end

      @mitemupdate.signal_connect( 'activate' ) do |w|
        @current.update
      end

      @mitemclear.signal_connect( 'activate' ) do |w|
        mbox = $CORE.mailset['trash']
        s = mbox.size
        s.times{|c| mbox.rm 0 }
        $CORE.notice "cleared #{s} mails from trash"
      end

      @mitemahead.signal_connect( 'activate' ) do |w|
        read_ahead
      end

      @mitemback.signal_connect( 'activate' ) do |w|
        read_back
      end

      @mitemnext.signal_connect( 'activate' ) do |w|
        @current.view_next
      end

      @mitemprev.signal_connect( 'activate' ) do |w|
        @current.view_prev
      end

      @mitemwrite.signal_connect( 'activate' ) do |w|
protect( 'write' ){    #D#
        $CORE.edit $CORE.composer.new_mail
        false
}    #D#
      end

      @mitemreply.signal_connect( 'activate' ) do |w|
protect( 'reply' ){   #D#
        orig = @current[ @current.selected ]
        unless orig then
          $CORE.notice "no mail selected to reply"
          break
        end
        $CORE.edit $CORE.composer.replyto( orig )
        false
}    #D#
      end

      @mitemreedit.signal_connect( 'activate' ) do |w|
protect( 're edit' ){   #D#
        orig = @current[ @current.selected ]
        unless orig then
          $CORE.notice "no mail selected to edit"
          break
        end
        $CORE.edit orig
}
      end

      @mitemfwd.signal_connect( 'activate' ) do |w|
protect( 'fwd' ){   #D#
        orig = @current[ @current.selected ]
        unless orig then
          $CORE.notice "no mail selected to forward"
          break
        end
        $CORE.edit $CORE.composer.forward( orig )
        false
}   #D#
      end

      @itemsend.signal_connect( 'activate' ) do |w|
protect( 'send' ){    #D#
        $CORE.sender.sendmail
        false
}    #D#
      end

      @itempop.signal_connect( 'activate' ) do |w|
protect( 'pop' ){    #D#
        # Thread.fork do
          $CORE.popper.load
        # end
        false
}    #D#
      end

      @itemquit.signal_connect( 'activate' ) do |w|
        $CORE.shutdown( false )
        false
      end

      @combo_pressing = false

      @mboxcombo.entry.signal_connect( 'changed' ) do |w|
protect( 'combo chang' ){    #D#
        unless @combo_pressing then
          chmbox w.get_text
        end
        false
}    #D#
      end

      @mboxcombo.button.signal_connect( 'pressed' ) do |w|
protect( 'combo press' ){    #D#
        @combo_pressing = true
        false
}    #D#
      end

      @mboxcombo.button.signal_connect( 'released' ) do |w|
protect( 'combo rel' ){   #D#
        @combo_pressing = false
        tmp = @mboxcombo.entry.get_text
        chmbox tmp
        false
}    #D#
      end
    end


    def nombox( name )
      nil
    end


    def realizembox( name, rc, lo )
      box = Gtk::HBox.new( false, 0 )
      adj = Gtk::Adjustment.new( 0, 0, 0, 0, 0, 0 )
      scr = Gtk::VScrollbar.new( adj )

      ifrc = $CORE.rc['if']['list']
      cli = Gtk::CList.new([ '', '',
        ifrc['date'], ifrc['from'], ifrc['subject']])
      cli.set_usize( 640, 170 )
      cli.set_column_width( 0, 16 )
      cli.set_column_width( 1, 16 )
      cli.set_column_width( 2, 110 )
      cli.set_column_width( 3, 200 )
      cli.set_vadjustment adj
      cli.selection_mode = Gtk::SELECTION_BROWSE

      box.pack_start scr, false, false, 0
      box.pack_start cli, true, true, 0

      @clistbox.pack_start( box, true, true, 0 )

      ret = RealMailbox.new( name, rc, lo,
        box, cli, adj, @mailpm, @mailmask, @mimepm, @mimemask )

      return ret
    end

  end   # class Mailset

  

  class Mailbox

    def initialize( name, rc, lo,
        base, cli, adj, pm_mail, msk_mail, pm_mime, msk_mime )
      super( name, rc, lo )

      @basewin = base
      @win = cli
      @adj = adj
      @mailpm   = pm_mail
      @mailmask = msk_mail
      @mimepm   = pm_mime
      @mimemask = msk_mime

      @clistsize = 0
      @selected = 0
      @unselected = false
      @multisel = false

      @win.signal_connect( 'select_row' ) do |w,row,col,eve|
        unless @multisel then
          viewmail row
        end
        false
      end

      @win.signal_connect( 'unselect_row' ) do |w,row,col,eve|
        @unselected = true
      end

      do_initialize
    end


    attr :selected
    attr :unselected

    def select( idx )
      if idx >= 0 and idx < size then
        @selected = idx
        @win.select_row idx, 0
      end
    end

    def each_selection
      @win.each_selection{|idx| yield idx }
    end


    def multiselect
      @multisel
    end

    def multiselect=( arg )
      @multisel = arg
      if @multisel then
        @win.selection_mode = Gtk::SELECTION_EXTENDED
      else
        @win.selection_mode = Gtk::SELECTION_BROWSE
      end
    end


    def view_next
      adjust_page @selected + 1
    end

    def view_prev
      adjust_page @selected - 1
    end

    def adjust_page( idx )
      return if idx < 0 or idx >= size

      val = [ idx * @win.row_height - @adj.page_size / 2, @adj.lower ].max
      upp = [ @adj.upper - @adj.page_size, @adj.lower ].max
      @adj.value = [ val, upp ].min
      select idx
    end
    private :adjust_page

    
    def repaint
      @win.freeze
      each_index do |i|
        # self[i].listcache = nil
        listmail i
      end
      @win.thaw

      select 0
    end
    

    def add( port )
      do_add( port )
      @win.append ['','','','','']
      listmail self.size - 1
    end

    #def cp( port )
      
    #def mv( port )

    def rm( idx )
      do_rm( idx )
      @win.remove_row idx
    end


    def show() @basewin.show_all end

    def hide() @basewin.hide_all end



    private


    def listmail( idx )
      port = self[idx]
      # unless arr = port.listcache then
        mail = Mail.new( port )
        # port.listcache =
        arr = $CORE.composer.list_items( mail )
      # end

      arr.each_with_index do |te,i|
        @win.set_text( idx, i + 2, te )
      end

      if port.unread then
        @win.set_pixmap( idx, 0, @mailpm, @mailmask )
      end
      if mail.multipart? then
        @win.set_pixmap( idx, 1, @mimepm, @mimemask )
      end
    end


    def viewmail( idx )
      return if idx >= self.size

      @selected = idx
      return unless @win.visible?
      port = self[idx]

      $CORE.notice ''
      port.unread = false
      @win.set_text( idx, 0, '' )

      $CORE.mailset.view Mail.new( port )
    end

  end   # class Mailbox



  class Popper

    def if_loggingin( client )
      @poppednum = 0
      super
      $CORE.notice "connecting #{client.address}..."
    end

    def if_loggedin( client )
      $CORE.notice "connected to #{client.address}"
      super
    end

    def if_popmail( port )
      super
      @poppednum += 1
      $CORE.notice "pop mail \##{@poppednum}"
    end

    def if_loggedout( client )
      super
      $CORE.notice "total #{@poppednum} mails popped."
    end
  
  end



  class Sender

    def if_nosendmail
      $CORE.notice "no mail to send"
    end

    def if_loggedin( client )
      $CORE.notice "connected to #{client.address}"
      @sentnum = 0
      super
    end

    def if_sendmail( mail, idx )
      super
      @sentnum += 1
      $CORE.notice "sent mail #{@sentnum}"
    end

    def if_loggedout( client )
      super
      $CORE.notice "sent total #{@sentnum} mails"
    end
  
  end



  class Composer

    def undecodable_mail( mail )
      $CORE.notice "cannot decode this mail"
      return super
    end

    def unknown_type( mail )
      $CORE.notice "cannot display this mail"
      return super
    end
  
  end
      


  class Editer

    def initialize( port, rc )
      port.must Port

      @port = port
      @port.ref
      @rc = rc
      @shuttingdown = false
      do_initialize
    end

    # for ruby level exit
    def shutdown( force = false )
      return if @shuttingdown
      @shuttingdown = true
      return unless force or shutdown_ok?
      edit_cancel
    end


    private

    def edit_ok
      i = mail2idx( @port )
      $CORE.mailset['draft'].select i
      $CORE.notice "edit draft\##{i}"
      in_shutdown
    end

    def edit_cancel
      i = mail2idx( @port )
      $CORE.mailset.rmmail 'draft', i
      $CORE.notice "discard draft\##{i}"
      in_shutdown
    end

    def in_shutdown
      @port.unref
      $CORE.edit_done self
    end


    def mail2idx( port )
      $CORE.mailset['draft'].index port
    end

  end   # class Editer



  class InternalEditer < Editer

    private

    def do_initialize
      @mail = Mail.new( @port )
      @buf = SizedStack.new

      set_keybind
      make_win
      pack_win
      init_event
      @win.show_all
    end


    def make_win
      mrc = @rc['menu']

      @win     = Gtk::Window.new( Gtk::WINDOW_TOPLEVEL )
      @win.set_title( 'mailest editer' )
      @win.border_width(0)
      @win.set_usize( 600, 700 )

      @box     = Gtk::VBox.new( false, 0 )
      #------------
      @butbox  = Gtk::HBox.new( false, 0 )

      @finbtn  = Gtk::Button.new( mrc['done'] )
      @finbtn.set_usize( 120, 22 )

      @intbtn  = Gtk::Button.new( mrc['cancel'] )
      @intbtn.set_usize( 120, 22 )
      #------------
      @tobox   = Gtk::HBox.new( false, 1 )

      @tolab   = Gtk::Label.new( mrc['to'] )
      @tolab.set_usize( 80, 22 )

      @toent   = Gtk::Entry.new
      @toent.set_usize( 460, 22 )
      @toent.set_editable true
      @toent.set_text @mail.to
      #------------
      @sbjbox  = Gtk::HBox.new( false, 1 )

      @sbjlab  = Gtk::Label.new( mrc['subject'] )
      @sbjlab.set_usize( 80, 22 )

      @sbjent  = Gtk::Entry.new
      @sbjent.set_usize( 460, 22 )
      @sbjent.set_editable true
      @sbjent.set_text @mail.subject
      #------------
      @textbox = Gtk::HBox.new( false, 0 )
      @adj     = Gtk::Adjustment.new(0,0,0,0,0,0)
      @scr     = Gtk::VScrollbar.new( @adj )

      # @text    = Gtk::Text.new( nil, @adj )
      @text    = Gtk::Buffer.new( nil, @adj )
      @text.insert_text( $CORE.composer.mail_to_viewtext(@mail), 0 )
      @text.set_editable true
    end

    def pack_win
      @win.add @box
      @box.pack_start( @butbox, false, false, 0 )
      @box.pack_start( @tobox, false, false, 0 )
      @box.pack_start( @sbjbox, false, false, 0 )
      @box.pack_start( @textbox, true, true, 0 )

      @butbox.pack_start( @finbtn, false, false, 0 )
      @butbox.pack_start( @intbtn, false, false, 0 )

      @tobox.pack_start( @tolab, false, false, 0 )
      @tobox.pack_start( @toent, true, true, 0 )

      @sbjbox.pack_start( @sbjlab, false, false, 0 )
      @sbjbox.pack_start( @sbjent, true, true, 0 )

      @textbox.pack_start( @text, true, true, 0 )
      @textbox.pack_start( @scr, false, false, 0 )
    end

    def init_event
      @finbtn.signal_connect( 'clicked' ) do |w|
protect( 'edit fin' ){   #D#
        edit_ok
}   #D#
      end

      @intbtn.signal_connect( 'clicked' ) do |w|
protect( 'interrupt' ){   #D#
        edit_cancel
}   #D#
      end

      @text.signal_connect( 'key_press_event' ) do |w, eve|
        c = (eve.state & Gdk::CONTROL_MASK != 0)
        a = (eve.state & Gdk::MOD1_MASK != 0)

        bind = nil
        if    c and a then bind = @ctrl_alt
        elsif c       then bind = @ctrl
        elsif       a then bind = @alt
        else               bind = @raw
        end
#p eve.keyval
#p bind[ eve.keyval ]
        if blk = bind[ eve.keyval ] then
          emitstop = blk.call( eve )
          @text.signal_emit_stop( 'key_press_event' ) if emitstop
        end
      end
    end


    def set_keybind
      @raw = {}
      @ctrl = {}
      @alt  = {}
      @ctrl_alt = {}

      @raw[Gdk::GDK_BackSpace] = proc{
# puts 'bs'
        # del nil, -1
				@text.bs
        true
      }
=begin
      @ctrl[Gdk::GDK_a] = @ctrl[Gdk::GDK_A] = proc{
puts 'atama'
        true
      }
=end
      @ctrl[Gdk::GDK_d] = @ctrl[Gdk::GDK_D] = proc{
# puts 'del'
        # del nil, 1
				@text.del
        true
      }
      @ctrl[Gdk::GDK_i] = @ctrl[Gdk::GDK_i] = proc{
# puts 'signature'
        File.open( File.expand_path '~/.signature' ) do |f|
          @text.insert_text f.read, @text.get_point
        end
        true
      }
=begin
      @ctrl[Gdk::GDK_k] = @ctrl[Gdk::GDK_K] = proc{
puts 'del line'
        pnt = @text.get_point
        if @text.get_chars( pnt, pnt + 1 ) == "\n" then
          del pnt, 1
          true
        else
          false
        end
      }
=end
      @ctrl[Gdk::GDK_s] = @ctrl[Gdk::GDK_S] = proc{
        edit_ok
        false
      }
      @ctrl[Gdk::GDK_q] = @ctrl[Gdk::GDK_Q] = proc{
        edit_cancel
        false
      }
      @ctrl[Gdk::GDK_u] = @ctrl[Gdk::GDK_U] = proc{
        cmd = @buf.pop and cmd.undo(@text)
        true
      }
      @ctrl[Gdk::GDK_y] = @ctrl[Gdk::GDK_Y] = proc{
        cmd = @buf.pop and cmd.yank(@text)
        true
      }
    end


    def del( from, len )
puts '- d ------------'
p from
p len
      from ||= @text.get_point
      if len < 0
        from += len
        len = -len
      end
p from
p len
      return if from < 0
      return if from + len >= @text.get_length

      # @text.freeze
      @text.set_point from
      str = @text.get_chars( from, from + len )
p "del=#{str}"
puts '---------------'
      @buf.push DelBackup.new( from, str )
      @text.forward_delete len
      # @text.thaw
    end


    def edit_ok
      @mail.body    = NKF.nkf( '-Ej', @text.get_chars( 0, @text.get_length ) )
      @mail.subject = NKF.nkf( '-Ej', @sbjent.get_text )
      @mail.to      = NKF.nkf( '-Ej', @toent.get_text )
      @mail.sync
      super
    end
  

    def edit_cancel
      @win.hide
      @win.destroy
      super
    end


    def shutdown_ok?
      true
    end

  end


  class SizedStack

    def initialize( up = 10 )
      @upper = up
      @arr = []
    end

    attr :upper, true

    def push( item )
      @arr.push item
      if @arr.size > @upper then
        @arr.shift
      end
    end

    def pop
      @arr.pop
    end

    def each
      @arr.each{|i| yield i }
    end

    def size
      @arr.size
    end

  end


  class Backup
  end

  class DelBackup < Backup
  
    def initialize( from, str )
      @from = from
      @str = str
    end

    def undo( text )
      text.insert_text @str, @from
    end

    def yank( text )
      text.insert_text @str, text.get_point
    end
  
  end

  class MultiBackup < Backup

    def initialize( *args )
      @args = args
    end

    def undo( text )
      @args.each{|cmd| cmd.undo( text ) }
    end

    def yank( text )
      @args.each{|cmd| cmd.yank( text ) }
    end
  
  end


  class ExternalEditer < Editer

    private


    def shutdown_ok?
      true
    end


    def do_initialize
      fname = sprintf( '/tmp/est%x_%x', $$, @port.id )
      File.open( fname, 'w' ) do |f|
        @port.ropen do |rp|
          while tmp = rp.read( 2048 )
            f.write tmp
          end
        end
      end

      Thread.fork do
        do_edit( fname )
      end
    end


    def do_edit( fname )
      stamp = File.mtime( fname )
      system "#{@rc['type']} #{fname}"
      if File.mtime( fname ) == stamp then   # not canceled
        edit_cancel
      else                                   # edited
        @port.wopen do |wp|
          File.open( fname ) do |f|
            while tmp = f.read( 2048 ) do
              wp.write tmp
            end
          end
        end
        edit_ok
      end
    end

  end

end   # module EST
