#
# addr.rb
#
#     Copyright(c) 1998-1999 Minero Aoki
#     aamine@dp.u-netsurf.ne.jp
#

require 'delegate'
require 'extmod'


module MailAddr ; end


class Mbox

  include MailAddr

  def initialize( adr, phr = '', ro = nil )
    self.addr   = adr
    self.phrase = phr
    @routes = []
    @routes.replace( ro.must Array ) if ro
  end


  property :addr,   String, true, true
  property :routes, Array,  true, false
  property :phrase, String, true, true


  def eql?( oth )
    self.addr   == oth.addr   and
    self.routes == oth.routes and
    self.phrase == oth.phrase
  end
  alias == eql?


  def to_s
    str = ''
    str << Bencode.encode( @phrase ) << ' ' if @phrase.size > 0
    return add_addr( str )
  end


  def inspect
    str = ''
    str += @phrase if @phrase.size > 0
    return add_addr( str )
  end

  
  def add_addr( str )
    ro = (@routes.size > 0)
    br = (str.size > 0) || ro

    str << '<' if br
    str << @routes.collect{|i| '@' + i }.join(',') << ':' if ro
    str << @addr
    str << '>' if br

    return str
  end
  private :add_addr

end   # class Mbox



class AddrGroup < DelegateClass( Array )

  def initialize( name, arg = nil )
    @name = (name.must String)

    if arg then
      super( arg.must(Array).dup )
    else
      super( [] )
    end
  end


  include Enumerable
  include MailAddr

  property :name, String, true, false

  
  def eql?( oth )
    super( oth ) and self.name == oth.name
  end

  def each_addr( &proc )
    each do |i|
      case i
      when AddrGroup then i.each_addr( &proc )
      when Mbox      then proc.call( i.to_s )
      else
        bug! "wrong type #{i.type} in addr group"
      end
    end
  end
  
  def to_s() @name + ':' << @items.join(', ') << ';' end

end   # class AddrGroup
