# mailest setup script

require 'ftools'
require 'rbconfig'
include Config


$interface = [
  'mailest'
]

$libdir = CONFIG['libdir'] + '/ruby/' + CONFIG['MAJOR'] + '.' + CONFIG['MINOR']
$bindir = CONFIG['bindir']
$estroot = $libdir + '/site_ruby/est'
$estlib  = $estroot

$conffile = <<EOS
require 'rbconfig'

module EST
  class Core
    def estconfig
      ret = {}

ret['rootdir'] = '#{$estroot}'
ret['libdir']  = '#{$estlib}'

      return ret
    end
    private :estconfig
  end
end
EOS


#---------------------------------------------------

def foreach_full( dn )
  dn = File.expand_path( dn )
  Dir.foreach( dn ) do |fn|
    f = dn + '/' + fn
    yield f if File.file? f
  end
end

def setup
  cur = Dir.pwd
  Dir.chdir 'strscan'
  load 'extconf.rb'
  system 'make'
  Dir.chdir cur

  File.open( 'est/estconf.rb', 'w' ) do |f|
    f.write $conffile
  end
end

def install
  puts "installing strscan..."
  cur = Dir.pwd
  Dir.chdir 'strscan'
  system 'make install'
  Dir.chdir cur

  puts "installing library scripts"
  foreach_full( 'lib' ) do |fn|
    File.install fn, $libdir
  end

  netdir = "#{$libdir}/net"
  File.mkpath netdir
  foreach_full( 'net' ) do |fn|
    File.install fn, netdir
  end

  puts "installing est system"
  File.mkpath $estlib
  foreach_full( 'est' ) do |fn|
    File.install fn, $estlib
  end

  puts "installing interfaces"
  $interface.each do |ifname|
    dn = "if/#{ifname}"
    iflib = $estroot + '/' + ifname
    File.mkpath iflib

    foreach_full( dn ) do |fn|
      if File.basename(fn) == ifname then
        File.install fn, iflib, 0755
        to = $bindir + '/' + ifname
        File.rm_f to
        File.symlink fn, to
      else
        File.install fn, iflib
      end
    end
  end

  print "\n\ndone all.\n\n\n"
end

#---------------------------------------------------

if not ARGV[0]             then setup
elsif ARGV[0] == 'install' then install
else
  puts 'usage:  setup.rb [install]'
end
